# -*- coding: utf-8; indent-tabs-mode: nil -*-

"""Facilitates the migration of users from one server to another.

More comment here...
"""

# These standard libs should all load fine
import sys
import os
import glob
import binascii
import time
import logging
import traceback
import threading
import getopt
import struct
import config

EMS_AB_ADDRESS_LOOKUP = 1

try:
    import threadpool
except ImportError:
    print "Threadpool library is missing."
    print "See http://www.chrisarndt.de/projects/threadpool for instructions on how"
    print "to install it."
    print "Or try: easy_install threadpool."
    print ""
    sys.exit(1)

try:
    import MAPI
    from MAPI.Util import *
except ImportError:
    print "The MAPI bindings appear to be absent. Make sure to install the"
    print "Zarafa python binding package."
    print ""
    sys.exit(1)

import string # moved here, since MAPI.Util contains 'string' variable

__author__ = "Mark Swaanenburg"
__copyright__ = "Copyright 2014, Zarafa"
__credits__ = ["Mark Swaanenburg", "Others"]
__license__ = "Proprietary"
__version__ = "0.1.0"
__maintainer__ = "Mark Swaanenburg"
__email__ = "m.swaanenburg@zarafa.com"
__status__ = "Beta/Unstable"


MUIDECSAB = DEFINE_GUID(0x50a921ac, 0xd340, 0x48ee, 0xb3, 0x19, 0xfb, 0xa7, 0x53, 0x30, 0x44, 0x25)
def DEFINE_ABEID(type, id):
    return struct.pack("4B16s3I4B", 0, 0, 0, 0, MUIDECSAB, 0, type, id, 0, 0, 0, 0)
EID_EVERYONE = DEFINE_ABEID(MAPI_DISTLIST, 1)


class NoSuchStoreError(Exception):
    def __init__(self): pass

class PathError(Exception):
    def __init__(self): pass

class ServerCollisionError(Exception):
    def __init__(self): pass

class Cancelled(Exception):
    def __init__(self): pass


def synchronized(lock):
    """ Synchronization decorator. """

    def wrap(f):
        def newFunction(*args, **kw):
            lock.acquire()
            try:
                return f(*args, **kw)
            finally:
                lock.release()
        return newFunction
    return wrap

def parseBool(v):
    return v.lower() in ("yes", "true", "t", "1")


# Opens the public by correctly creating a provider in the profile, and then opening it. This makes sure
# that the public store is openend as an actual public store, and not as a private store
def GetPublic(session, entryid):
    sa = session.AdminServices(0)
    table = sa.GetMsgServiceTable(0)
    table.Restrict(SPropertyRestriction(RELOP_EQ, PR_SERVICE_NAME, SPropValue(PR_SERVICE_NAME, 'ZARAFA6')), 0)
    table.SetColumns([PR_SERVICE_UID], 0)
    serviceUid = table.QueryRows(1, 0)[0][0].Value

    pa = sa.AdminProviders(serviceUid, 0)

    # Create a new public provider
    newprov = pa.CreateProvider('ZARAFA6_MSMDB_public', [], 0, 0)

    # Open the profile section for the public provider
    ps = pa.OpenProfileSection(newprov, None, 0)

    ps.SetProps([SPropValue(PR_ENTRYID, entryid)])

    store = session.OpenMsgStore(0, entryid, IID_IMsgStore, MAPI_BEST_ACCESS)

    return store

class Logger(ECSimpleLogger):
    def __init__(self, logger):
        super(Logger,self).__init__()
        self.logger = logger

    def Log(self, loglevel, message):
        self.logger.debug(message)

class UserContext:
    """Encapsulates all information about a user store on a particular server."""

    def __init__(self, session, store, user):
        self.session = session
        self.store = store
        self.root = self.store.OpenEntry(None, IID_IMAPIFolder, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS)
        self.user = user
        self.logger = logging.getLogger("UserContext[%s]" % self.user.replace('.', '_'))
        self.syncLogger = Logger(self.logger)

    def __getattr__(self, attr):
        if attr == 'StoreEntryID':
            return self.store.GetProps([PR_ENTRYID], 0)[0].Value
        elif attr == 'RootEntryID':
            return self.root.GetProps([PR_ENTRYID], 0)[0].Value
        elif attr == 'StoreGUID':
            return self.store.GetProps([PR_RECORD_KEY], 0)[0].Value
        elif attr == 'UserEntryID':
            if self.user == '__public__':
                return EID_EVERYONE
            return self.store.GetProps([PR_MAILBOX_OWNER_ENTRYID], 0)[0].Value
        elif attr == 'DisplayName':
            return self.store.GetProps([PR_DISPLAY_NAME], 0)[0].Value
        elif attr == 'SyncRoot':
            return self.root
        elif attr == 'SyncRootEntryID':
            return self.SyncRoot.GetProps([PR_ENTRYID], 0)[0].Value
        elif attr == 'Hierarchy':
            def gen():
                table = self.SyncRoot.GetHierarchyTable(CONVENIENT_DEPTH)
                table.SetColumns([PR_ENTRYID, PR_FOLDER_TYPE], TBL_BATCH)
                while True:
                    rows = table.QueryRows(16, 0)
                    if len(rows) == 0:
                        break
                    for row in rows:
                        if row[1].Value == FOLDER_GENERIC:
                            yield self.root.OpenEntry(row[0].Value, IID_IMAPIFolder, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS)
            return gen()
        elif attr == 'User':
            return self.user
        elif attr == 'IsPublic':
            mdb_provider = self.store.GetProps([PR_MDB_PROVIDER], 0)[0]
            return mdb_provider.ulPropTag == PR_MDB_PROVIDER and mdb_provider.Value == ZARAFA_STORE_PUBLIC_GUID;
        elif attr == 'ServerGUID':
            return self.store.GetProps([0x67c20102], 0)[0].Value
        else:
            raise AttributeError()

    def __repr__(self):
        return "%s: %s - %s" % (self.DisplayName, binascii.hexlify(self.StoreEntryID), binascii.hexlify(self.RootEntryID))

    def Log(self, loglevel, message):
        self.logger.debug(message)

    def SyncHierarchyTo(self, other, streamstorage):
        """Synchronize the hierarchy from this contexts store to another.

        other: The UserContext containing the store to synchronize the hierarchy to.
        streamstorage: A PathBasedStreamStorage object that contains the state streams.
        """
        syncprops = [
            PR_ENTRYID,
            PR_CONTAINER_CLASS,
            PR_DISPLAY_NAME,
            PR_COMMENT,
            PR_IPM_APPOINTMENT_ENTRYID,
            PR_IPM_CONTACT_ENTRYID,
            PR_IPM_DRAFTS_ENTRYID,
            PR_IPM_JOURNAL_ENTRYID,
            PR_IPM_NOTE_ENTRYID,
            PR_IPM_TASK_ENTRYID,
            PR_REM_ONLINE_ENTRYID,
            PR_ADDITIONAL_REN_ENTRYIDS,
            PR_FREEBUSY_ENTRYIDS,
            PR_RULES_DATA,
            PR_FOLDER_XVIEWINFO_E,
            PR_FOLDER_DISPLAY_FLAGS,
            PR_IPM_OL2007_ENTRYIDS,
            PR_ACL_DATA
        ]

        exporter = self.SyncRoot.OpenProperty(PR_HIERARCHY_SYNCHRONIZER, IID_IExchangeExportChanges, 0, 0)
        importer = other.SyncRoot.OpenProperty(PR_COLLECTOR, IID_IExchangeImportHierarchyChanges, 0, 0)
        stream = streamstorage.LoadStateStream(self.ServerGUID + self.SyncRootEntryID)
        importstream = streamstorage.LoadStateStream(other.ServerGUID + other.SyncRootEntryID)

        importer.Config(importstream, 0)
        exporter.Config(stream, SYNC_NORMAL|SYNC_UNICODE, importer, None, syncprops, None, 64)

        ex = exporter.QueryInterface(IID_IECExportChanges)
        ex.SetLogger(self.syncLogger)

        self.logger.debug("Starting hierarchy synchronization")
        step = 0
        retry = 0
        while True:
            try:
                (steps, step) = exporter.Synchronize(step)
                retry = 0
                self.logger.debug("Synchronized folder %d from %d" % (step, steps))
                if (steps == step):
                    break;
            except MAPIError, err:
                if err.hr in [MAPI_E_NETWORK_ERROR, MAPI_E_TIMEOUT, MAPI_E_CALL_FAILED, MAPI_E_DISK_ERROR]:
                    if retry < 5:
                        logging.fatal("(%d) Received a network error or timeout." % (retry))
                    else:
                        raise
                    retry += 1
                elif err.hr == MAPI_E_NOT_FOUND:
                    logging.fatal("The synchronization state for this store is no longer valid. Remove the destination store and re-migrate all the data for this user.")
                    raise
                else:
                    raise
                logging.fatal("continuing relocation")

        exporter.UpdateState(stream)
        importer.UpdateState(importstream)
        streamstorage.SaveStateStream(self.ServerGUID + self.SyncRootEntryID, stream)
        streamstorage.SaveStateStream(other.ServerGUID + other.SyncRootEntryID, importstream)

        self.logger.debug("Hierarchy synchronized")

    def CopyReceiveFoldersTo(self, other):
        """Sets the receive folders of the other contexts store based on those of the current.

        other: The UserContext containing the other store.
        """
        self.logger.debug("Copying receive folders")
        try:
            table = self.store.GetReceiveFolderTable(0)
            table.SetColumns([PR_MESSAGE_CLASS,PR_ENTRYID], 0)
            rows = table.QueryRows(16, 0)
            for row in rows:
                try:
                    self.logger.debug("Copying receive folder for class '%s': %s" % (row[0].Value, row[1].Value.encode('hex')))
                    other.store.SetReceiveFolder(row[0].Value, 0, row[1].Value)
                except MAPIError, err:
                    self.logger.warning("Failed to set receive folder for message class '%s' (err: %s)" % (row[0].Value, err))
            self.logger.debug("Receive folders copied")
        except MAPIError, err:
            if err.hr != MAPI_E_NO_SUPPORT:
                raise
            self.logger.debug("Store does not support receive folders")

    def CopyProps(self, fromObj, toObj):
        tags = fromObj.GetPropList(MAPI_UNICODE)
        tags.append(PR_EC_PUBLIC_IPM_SUBTREE_ENTRYID)   # Only does something for the public store
        tags.append(PR_ACL_DATA)                        # Hidden store property
        props = fromObj.GetProps(tags, 0)

        # Copy all the small properties
        try:
            toObj.SetProps(props)
        except MAPIError:
            pass

        # Go through the properties to copy the large properties
        for prop in zip(tags, props):
            if PROP_TYPE(prop[1].ulPropTag) == PT_ERROR and prop[1].Value == MAPI_E_NOT_ENOUGH_MEMORY:
                try:
                    fromStream = fromObj.OpenProperty(prop[0], IID_IStream, 0, 0)
                    toStream = toObj.OpenProperty(prop[0], IID_IStream, 0, MAPI_CREATE|MAPI_MODIFY)
                    fromStream.CopyTo(toStream, 0xffffff)
                except MAPIError:
                    pass


    def CopyUnsyncablePropsTo(self, other):
        """Copies 'unsyncable' properties from the current context to another.

        other: The UserContext to sync the properties to.

        The 'unsyncable' properties are those found in the store and the root container.
        """
        self.logger.debug("Copying store properties")
        self.CopyProps(self.store, other.store)

        self.logger.debug("Copying root properties")
        self.CopyProps(self.root, other.root)

    def GetChangeAdvisor(self):
        """Get a change advisor for the current context.
        """
        return self.store.OpenProperty(PR_EC_CHANGE_ADVISOR, IID_IECChangeAdvisor, 0, 0)

    def GetFolder(self, entryID):
        """Get the folder specified by the passed entryid.

        entryID: The entryid that specifies the folder to open.
        """
        return self.store.OpenEntry(entryID, IID_IMAPIFolder, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS)

    def GetImporterFor(self, entryID):
        """ Get an change importer for the folder specified by the passed entryid.

        entryID: The enrtyid for the folder to get the change importer for.
        """
        folder = self.store.OpenEntry(entryID, IID_IMAPIFolder, MAPI_BEST_ACCESS|MAPI_DEFERRED_ERRORS)
        importer = folder.OpenProperty(PR_COLLECTOR, IID_IExchangeImportContentsChanges, 0, 0)
        return importer




class AdminSession:
    lock = threading.Lock()     # static lock object (shared between instances)

    @synchronized(lock)
    def __init__(self, server = None, **keywords):
        self.forceServer = parseBool(keywords.pop('force_source', 'no'))

        if server is None:
            server = os.getenv('ZARAFA_SOCKET')

        if server.startswith('file://'): # Always use SYSTEM for file:// URLs since that will always work anyway
            keywords['username'] = 'SYSTEM'
            keywords['password'] = ''

        keywords['flags'] = 0
        try:
            self.session = OpenECSession(keywords.pop('username', 'SYSTEM'), keywords.pop('password', ''), server, **keywords)
            self.store = GetDefaultStore(self.session);
            self.logger = logging.getLogger("AdminSession[%s]" % server.replace('.', '_'))
        except MAPIError, e:
            logging.fatal("Unable to open connection to server '%s'" % (server))
            raise e

    def GetUserContext(self, user, default = True):
        try:
            if user == '__public__':
                flags = 0
                if not default or self.forceServer:
                    flags = EC_OVERRIDE_HOMESERVER

                sa = self.store.QueryInterface(IID_IECServiceAdmin)
                storeEid = sa.GetPublicStoreEntryID(flags)
                store = GetPublic(self.session, storeEid)
            else:
                flags = 0
                if not default or self.forceServer:
                    # OPENSTORE_OVERRIDE_HOME_MDB
                    flags = 0x10

                ema = self.store.QueryInterface(IID_IExchangeManageStore)
                storeEid = ema.CreateStoreEntryID(None, user, flags)
                store = self.session.OpenMsgStore(0, storeEid, IID_IMsgStore, MAPI_BEST_ACCESS)

            return UserContext(self.session, store, user)

        except MAPIError, err:
            if err.hr == MAPI_E_NOT_FOUND:
                self.logger.debug("Store for user '%s' not found" % user)
                raise NoSuchStoreError()
            self.logger.error("Failed to open store for user '%s'" % user)
            raise

    # Resolve user EID
    def GetUserEID(self, user):
        if user == '__public__':
            return EID_EVERYONE

        ab = self.session.OpenAddressBook(0, None, 0)
        gabid = ab.GetDefaultDir()
        gab = ab.OpenEntry(gabid, None, 0)
        (result, flags) = gab.ResolveNames([PR_DISPLAY_NAME, PR_ENTRYID], EMS_AB_ADDRESS_LOOKUP, [[SPropValue(PR_DISPLAY_NAME, user)]], [MAPI_UNRESOLVED])
        return result[0][1].Value

    def ExtractStoreGuid(self, storeId):
        id = storeId[22:]               # Remove AB, GUID and 2 flag bytes
        id = id.split('\000', 1)[1]     # Remove provider dll
        return id[4:20]                 # Get the store GUID

    def GetStoreOwner(self, storeId):
        guid = self.ExtractStoreGuid(storeId)
        if guid is None:
            self.logger.error("Unable to extract GUID from store identifier")
            return (None, None)
        sa = self.store.QueryInterface(IID_IECServiceAdmin)
        table = sa.OpenUserStoresTable(0)
        table.SetColumns([PR_EC_USERNAME], TBL_BATCH)
        try:
            table.FindRow(SPropertyRestriction(RELOP_EQ, PR_EC_STOREGUID, SPropValue(PR_EC_STOREGUID, guid)), BOOKMARK_BEGINNING, 0)
        except MAPIError, er:
            if er.hr != MAPI_E_NOT_FOUND:
                raise
            return (guid, None)
        rows = table.QueryRows(1, 0)
        if len(rows) == 0:
            return (guid, None)     # Quite broken if this happens.
        if PROP_TYPE(rows[0][0].ulPropTag) == PT_ERROR:
            return (guid, None)
        return (guid, rows[0][0].Value)


    def CloneStoreFrom(self, srcContext):
        sa = self.store.QueryInterface(IID_IECServiceAdmin)
        store_type = ECSTORE_TYPE_PRIVATE
        if srcContext.IsPublic:
            store_type = ECSTORE_TYPE_PUBLIC
        try:
            sa.CreateEmptyStore(store_type, self.GetUserEID(srcContext.user), EC_OVERRIDE_HOMESERVER, srcContext.StoreEntryID, srcContext.RootEntryID)
        except MAPIError, ex:
            if ex.hr == MAPI_E_COLLISION:
                self.logger.error("Unable to create the store: a store with the same identifier already exists on the destination server.")
                guid, owner = self.GetStoreOwner(srcContext.StoreEntryID)
                if not guid is None:
                    if owner is None:
                        self.logger.error("Current store is orphaned, guid: %s" % guid.encode('hex'))
                    else:
                        self.logger.error("Current store owner: '%s'" % owner)
                raise
        return self.GetUserContext(srcContext.User, False)

    def CloneReceiveFolders(self, srcContext, dstContext, streamstorage):
        # We can't set receivefolders to folders that don't exist. So we need to sync the hierarchy first.
        srcContext.SyncHierarchyTo(dstContext, streamstorage)
        srcContext.CopyReceiveFoldersTo(dstContext)
        srcContext.CopyUnsyncablePropsTo(dstContext)



class Migrator:
    def __init__(self, args, streamstorage):
        self.user = args['user']
        self.target = args['server']
        self.streamstorage = streamstorage
        self.advisesink = args['sinkmonitor'].GetChangeAdviseSink(self, False)
        self.advisesinkother = args['sinkmonitor'].GetChangeAdviseSink(self, True)
        self.exitcontrol = args['exitcontrol']
        self.connection = args['connection']
        self.relations = {}
        self.logger = logging.getLogger("Migrator[%s]" % self.user.replace('.', '_'))
        self.syncLogger = Logger(self.logger)
        self.lock = threading.Lock()

    def EnableMonitor(self):
        self.advisesink.EnableMonitor()
        self.advisesinkother.EnableMonitor()

    def LoadStateStream(self, entryid, overwrite = False):
        return self.streamstorage.LoadStateStream(entryid, overwrite)

    def SaveStateStream(self, entryid, stream):
        self.streamstorage.SaveStateStream(entryid, stream)

    def ResetState(self, s1guid, s2guid, storeguid):
        self.streamstorage.ResetState(s1guid, s2guid, storeguid)

    def Initialize(self):
        self.logger.debug("Opening default admin session")
        srcAdminSession = AdminSession(self.connection.get('serverpath', 'file:///var/run/zarafad/server.sock'), **self.connection)

        self.logger.debug("Opening admin session to '%s'" % self.target)
        dstAdminSession = AdminSession(self.target, **self.connection)

        self.logger.debug("Opening source store")
        srcContext = srcAdminSession.GetUserContext(self.user)
        dstContext = None
        try:
            # Resolve user store without redirection
            self.logger.debug("Opening destination store")
            dstContext = dstAdminSession.GetUserContext(self.user, False)
            self.streamstorage.SetServerGuids(srcContext.ServerGUID, dstContext.ServerGUID)
        except NoSuchStoreError:
            # Create a new store on the target server.
            self.logger.debug("Opening destination store failed")
            try:
                self.logger.debug("Cloning store")
                dstContext = dstAdminSession.CloneStoreFrom(srcContext)
                self.streamstorage.SetServerGuids(srcContext.ServerGUID, dstContext.ServerGUID)
                self.logger.debug("Resetting state")
                self.ResetState(srcContext.ServerGUID, dstContext.ServerGUID, srcContext.StoreGUID)
                self.logger.debug("Cloning receivefolders")
                dstAdminSession.CloneReceiveFolders(srcContext, dstContext, self.streamstorage)
            except:
                self.logger.error("Store creation failed")
                raise

        if srcContext.ServerGUID == dstContext.ServerGUID:
            raise ServerCollisionError()

        # Update migrator state
        self.srcContext = srcContext
        self.dstContext = dstContext

    def SynchronizeFolder(self, srcFolder, srcContext, dstContext, other):
        includeProps = [
            PR_ENTRYID,
            PR_SOURCE_KEY,
            PR_EC_IMAP_ID,
            PR_EC_IMAP_SUBSCRIBED,
            PR_EC_IMAP_MAX_ID,
            PR_EC_IMAP_EMAIL,
            PR_EC_IMAP_EMAIL_SIZE,
            PR_EC_IMAP_BODY,
            PR_EC_IMAP_BODYSTRUCTURE
        ]

        foldername = srcFolder.GetProps([PR_DISPLAY_NAME], 0)[0].Value
        self.logger.debug("Synchronizing folder: %s" % foldername)
        entryID = srcFolder.GetProps([PR_ENTRYID], 0)[0].Value

        stream = self.LoadStateStream(srcContext.ServerGUID + entryID)
        importstream = self.LoadStateStream(dstContext.ServerGUID + entryID)

        try:
            self.logger.debug("Get importer for %s" % (entryID.encode('hex')))
            importer = dstContext.GetImporterFor(entryID)
            importer.Config(importstream, 0)
            exporter = srcFolder.OpenProperty(PR_CONTENTS_SYNCHRONIZER, IID_IExchangeExportChanges, 0, 0)
            exporter.Config(stream, SYNC_NORMAL | SYNC_ASSOCIATED | SYNC_READ_STATE, importer, None, includeProps, None, 64)

            ex = exporter.QueryInterface(IID_IECExportChanges)
            ex.SetLogger(self.syncLogger)

        except:
            logging.exception('Unable to sync')
            raise

        try:
            step = 0
            retry = 0
            while True:
                    if self.exitcontrol.exit_pending:
                        raise Cancelled()
                    steps, step = exporter.Synchronize(step);
                    retry = 0
                    self.logger.debug("%s: %u / %u" % (foldername, step, steps))
                    if steps == step:
                        break

        finally:
            exporter.UpdateState(stream)
            importer.UpdateState(importstream)

            self.SaveStateStream(srcContext.ServerGUID + entryID, stream)   # Persist the state stream
            self.SaveStateStream(dstContext.ServerGUID + entryID, importstream)

        # Make sure to map the syncid to the entryID.
        state = SyncState(stream)
        self.relations[srcContext.ServerGUID + str(state.syncid)] = srcFolder
        # Make sure the change advisor is monitoring the folder
        if other:
            advisor = self.advisorOther
        else:
            advisor = self.advisor
        try:
            advisor.IsMonitoringSyncId(state.syncid)
        except MAPIError:
            self.logger.debug("Adding syncstate '%s' to change advisor" % state)
            advisor.AddKeys([state.key])

        return state

    def SynchronizeFolderById(self, syncid, other):
        if other:
            guid = self.dstContext.ServerGUID
        else:
            guid = self.srcContext.ServerGUID

        logging.debug('Sync by folderid for syncid %s, other %s' % (syncid, str(other)))
        if self.relations.has_key(guid + str(syncid)):
            srcFolder = self.relations[guid + str(syncid)]
            retry = 0
            while True:
                try:
                    if other:
                        self.advisesinkother.UpdateSyncState(self.SynchronizeFolder(srcFolder, self.dstContext, self.srcContext, other))
                    else:
                        self.advisesink.UpdateSyncState(self.SynchronizeFolder(srcFolder, self.srcContext, self.dstContext, other))
                    break
                except MAPIError, err:
                    if err.hr in [MAPI_E_NETWORK_ERROR, MAPI_E_TIMEOUT, MAPI_E_CALL_FAILED, MAPI_E_DISK_ERROR]:
                        if retry < 5:
                            self.logger.fatal("(%d) Received a network error or timeout (%x)." % (retry, err.hr))
                        else:
                            self.logger.error("Too many errors, aborting")
                            raise
                        retry += 1
                    else:
                        raise
                    self.logger.fatal("continuing relocation")

    def SyncOneWay(self):
        if self.exitcontrol.exit_pending:
            raise Cancelled()

        # Get stream for change notifications
        stream = None

        self.advisor = self.srcContext.GetChangeAdvisor()
        self.advisor.Config(stream, None, self.advisesink, 0)

        try:
            self.srcContext.SyncHierarchyTo(self.dstContext, self.streamstorage)
        except MAPIError, e:
            self.logger.error("Hierarchy migration failed: %x" % e.hr)
            raise

        try:
            for folder in self.srcContext.Hierarchy:
                retry = 0
                while True:
                    try:
                        self.SynchronizeFolder(folder, self.srcContext, self.dstContext, False)
                        break
                    except MAPIError, err:
                        if err.hr in [MAPI_E_NETWORK_ERROR, MAPI_E_TIMEOUT, MAPI_E_CALL_FAILED, MAPI_E_DISK_ERROR]:
                            if retry < 5:
                                self.logger.fatal("(%d) Received a network error or timeout. (%x)" % (retry, err.hr))
                            else:
                                self.logger.error("Too many errors, aborting")
                                raise
                            retry += 1
                        else:
                            raise
                        self.logger.fatal("continuing relocation")

        except MAPIError, e:
            self.logger.error("Contents migration failed: %x" % e.hr)
            raise

        self.srcContext.CopyReceiveFoldersTo(self.dstContext)
        self.srcContext.CopyUnsyncablePropsTo(self.dstContext)


    def SyncOtherWay(self):
        if self.exitcontrol.exit_pending:
            raise Cancelled()

        # Get stream for change notifications
        stream = None

        self.advisorOther = self.dstContext.GetChangeAdvisor()
        self.advisorOther.Config(stream, None, self.advisesinkother, 0)

        self.dstContext.SyncHierarchyTo(self.srcContext, self.streamstorage)
        [self.SynchronizeFolder(folder, self.dstContext, self.srcContext, True) for folder in self.dstContext.Hierarchy]

def ReadUInt(stream):
    """Read one UInt from a binary stream.
    """
    return struct.unpack("I", stream.Read(4))[0]


class SyncState:
    """Represent a syncstate.

    A syncstate is defined as a four byte syncid and a four byte change id concatenated
    in a binary stream.
    """
    def __init__(self,arg):
        """Initialize a SyncState object

        arg: Can be an IStream object containing at least eight bytes or a struct
             containing the syncid and changeid.
        """
        if type(arg) == type(IStream()):
            clone = arg.Clone()
            clone.Seek(0, BOOKMARK_BEGINNING)
            self.syncid = ReadUInt(clone)
            self.changeid = ReadUInt(clone)
        else:
            self.syncid, self.changeid = struct.unpack("II", arg)

    def __getattr__(self, attr):
        if attr == 'syncid':
            return self.syncid
        elif attr == 'changeid':
            return self.changeid
        elif attr == 'key':
            return struct.pack("II", self.syncid, self.changeid)
        else:
            raise AttributeError

    def __repr__(self):
        return "SyncState: syncid=%u, changeid=%u" % (self.syncid, self.changeid)



class AdviseSink(ECChangeAdviseSink):
    def __init__(self, migrator, collection, other):
        ECChangeAdviseSink.__init__(self, [IID_IECChangeAdviseSink])
        self.migrator = migrator
        self.collection = collection
        self.lock = threading.Lock()
        self.monitored = False
        self.events = {}
        self.other = other;

    def OnNotify(self, flags, entries):
        if len(entries) > 0:
            states = (SyncState(entry) for entry in entries)
            newevents = dict([state.syncid, state.changeid] for state in states)
            self.lock.acquire()
            try:
                if self.monitored == True:
                    self.collection.AddEvents(self, newevents)
                else:
                    self.events.update(newevents)
            finally:
                self.lock.release()
        return 0

    def EnableMonitor(self):
        self.lock.acquire()
        try:
            if (self.monitored == False):
                self.monitored = True
                self.collection.AddEvents(self, self.events)
                self.events = None
        finally:
            self.lock.release()

    def UpdateSyncState(self, state):
        self.collection.UpdateState(self, state)

    def SynchronizeFolder(self, syncid):
        # Use a lock to make sure a Migrator isn't used on multiple worker threads simultaneously.
        self.migrator.lock.acquire()
        try:
            retry = 0
            while True:
                try:
                    self.migrator.SynchronizeFolderById(syncid, self.other)
                    break

                except MAPIError, err:
                    if err.hr in [MAPI_E_NETWORK_ERROR, MAPI_E_TIMEOUT, MAPI_E_CALL_FAILED, MAPI_E_DISK_ERROR]:
                        if retry < 5:
                            logging.fatal("(%d) Received a network error or timeout." % (retry))
                        else:
                            logging.error("Too many errors, aborting")
                            raise
                        retry += 1
                    else:
                        raise
                    logging.fatal("continuing relocation")
        finally:
            self.migrator.lock.release()


class AdviseSinkMonitor:
    def __init__(self):
        self.condition = threading.Condition()
        self.doExit = False
        self.sinkevents = {}
        self.logger = logging.getLogger("AdviseSinkMonitor")

    def GetChangeAdviseSink(self, migrator, other):
        return AdviseSink(migrator, self, other)

    def AddEvents(self, sink, newevents):
        self.logger.debug("Adding %u new events from sink %s" % (len(newevents), sink.__hash__))
        self.condition.acquire()
        try:
            if (self.sinkevents.has_key(sink)):
                events = self.sinkevents[sink]
            else:
                self.sinkevents[sink] = events = {}
            events.update(newevents)
            self.condition.notify()
        finally:
            self.condition.release()

    def UpdateState(self, sink, state):
        self.logger.debug("Updating state for sink %s, syncid=%u, changeid=%u" % (sink.__hash__, state.syncid, state.changeid))
        self.condition.acquire()
        try:
            events = self.sinkevents.get(sink, None)
            if events and events.has_key(state.syncid) and events[state.syncid] <= state.changeid:
                del events[state.syncid]
                if not events:
                    del self.sinkevents[sink]
        finally:
            self.condition.release()

    def Cancel(self):
        self.logger.debug("Canceling AdviseSinkMonitor...")
        self.condition.acquire()
        try:
            self.doExit = True
            self.condition.notify()
        finally:
            self.condition.release()

    def Run(self, pool):
        self.logger.debug("AdviseSinkMonitor started")
        self.condition.acquire()
        try:
            while not self.doExit:
                while not self.doExit and not self.sinkevents:
                    self.logger.debug("Waiting for events...")
                    self.condition.wait()
                if not self.doExit:
                    self.logger.debug("Get events for %u sinks" % len(self.sinkevents))
                    requests = threadpool.makeRequests(self.ProcessSink, ([item] for item in self.sinkevents.items()))
                    self.sinkevents = {}
                    [pool.putRequest(req) for req in requests]
        finally:
            self.condition.release()

    def ProcessSink(self, args):
        sink, events = args[0]
        self.logger.debug("Processing %u events for sink %s" % (len(events), sink.__hash__))
        [sink.SynchronizeFolder(syncid) for syncid in events.keys()]




class PathBasedStreamStorage:
    def __init__(self, basepath):
        self.logger = logging.getLogger("PathBasedStreamStorage")
        self.basepath = os.path.abspath(basepath)
        if os.path.exists(self.basepath) and not os.path.isdir(self.basepath):
            self.logger.critical("'%s' exists but is not a directory!" % self.basepath)
            raise PathError()
        elif not os.path.exists(self.basepath):
            self.logger.debug("Creating '%s'" % self.basepath)
            os.mkdir(self.basepath)
        self.curpath = None

    def SetServerGuids(self, srcGuid, dstGuid):
        self.logger.debug("srcGuid=%s" % binascii.hexlify(srcGuid))
        self.logger.debug("dstGuid=%s" % binascii.hexlify(dstGuid))
        self.curpath = '%s/%s-%s' % (self.basepath, binascii.hexlify(srcGuid), binascii.hexlify(dstGuid))
        if os.path.exists(self.curpath) and not os.path.isdir(self.curpath):
            self.logger.critical("'%s' exists but is not a directory!" % self.curpath)
            raise PathError()
        elif not os.path.exists(self.curpath):
            self.logger.debug("Creating '%s'" % self.curpath)
            os.mkdir(self.curpath)

    def LoadStateStream(self, entryid, overwrite = False):
        self.logger.debug("Loading state for '%s'" % binascii.hexlify(entryid))
        filename = '%s/%s.state' % (self.curpath, binascii.hexlify(entryid))
        stream = IStream()
        stream.Write('0000000000000000'.decode('hex'))

        if not overwrite:
            try:
                f = open(filename, 'rb')
                bin = f.read()
                stream = IStream()
                stream.Write(bin)
            except IOError:
                pass
        return stream

    def ResetState(self, server1guid, server2guid, storeguid):
        states = glob.glob('%s/%s00000000%s*' % (self.curpath, binascii.hexlify(server1guid), binascii.hexlify(storeguid)))
        states += glob.glob('%s/%s00000000%s*' % (self.curpath, binascii.hexlify(server2guid), binascii.hexlify(storeguid)))

        self.logger.debug('%s/%s00000000%s*' % (self.curpath, binascii.hexlify(server1guid), binascii.hexlify(storeguid)))

        for state in states:
            os.unlink(state)

        self.logger.debug("Removed %d states" % (len(states)))

    def SaveStateStream(self, entryid, stream):
        self.logger.debug("Saving state for '%s'" % binascii.hexlify(entryid))
        filename = '%s/%s.state' % (self.curpath, binascii.hexlify(entryid))
        clone = stream.Clone()
        clone.Seek(BOOKMARK_BEGINNING, 0)
        bin = clone.Read(0xffffff)
        f = open(filename, 'wb')
        f.write(bin)



class ExitControl:
    def __init__(self):
        self.doExit = False
        self.logger = logging.getLogger("ExitControl")

    def initiate_exit(self):
        self.logger.debug("Exit initiated...")
        self.doExit = True

    def __getattr__(self, attr):
        if attr == 'exit_pending':
            return self.doExit
        else:
            raise AttributeError()



class ProcessTracker:
    def __init__(self):
        self.lock = threading.Lock()
        self.started = 0
        self.completed = 0
        self.failed = 0
        self.canceled = 0

    def __getattr__(self, attr):
        if attr == 'finished':
            return self.completed + self.failed + self.canceled
        elif attr == 'completed':
            return self.completed
        elif attr == 'failed':
            return self.failed
        elif attr == 'canceled':
            return self.canceled
        else:
            raise AttributeError()

    def start_migration(self, args):
        self.lock.acquire()
        try:
            self.started += 1
        finally:
            self.lock.release()

        # check for exit before we print that we start the next user to avoid confusion
        if args['exitcontrol'].exit_pending:
            raise Cancelled()

        ss = PathBasedStreamStorage(os.getenv('HOME') + '/.zarafa-migrate-store')
        logging.info("Starting to relocate user '%s' to '%s'" % (args['user'], args['server']))

        while True:
            try:
                m = Migrator(args, ss)
                m.Initialize()
                m.SyncOneWay()
                if args['bidirectional']:
                    m.SyncOtherWay()
            except MAPIError, er:
                if er.hr == MAPI_E_NO_SUPPORT:
                    logging.error("Unable to syncronize, a server may have 'enable_enhanced_ics = no' in its configuration file")
                else:
                    logging.error("Unexpected MAPI error 0x%08X" % er.hr)
                logging.error("Migration failed")
                raise
            except:
                logging.error("Migration failed")
                raise

            logging.info("Maintaining sync for user '%s'..." % args['user'])
            m.EnableMonitor()
            break


    def migration_completed(self, request, result):
        self.lock.acquire()
        try:
            self.completed += 1
        finally:
            self.lock.release()

        logging.info("Initial migration completed for user '%s'" % request.args[0]['user'])

    def migration_exception(self, request, exc_info):
        if exc_info[0] is Cancelled:
            self.lock.acquire()
            try:
                self.canceled += 1
            finally:
                self.lock.release()

            logging.info("Migration for user '%s' was canceled" % request.args[0]['user'])
        else:
            self.lock.acquire()
            try:
                self.failed += 1

                logging.error("Migration for user '%s' failed" % request.args[0]['user'])
                lines = traceback.format_exception(exc_info[0], exc_info[1], exc_info[2])
                for line in lines:
                    [logging.debug("exc_info: %s" % s) for s in string.split(string.rstrip(line, '\n'), '\n')]
            finally:
                self.lock.release()


def main(argv = None):
    if argv is None:
        argv = sys.argv

    try:
        opts, args = getopt.gnu_getopt(argv[1:], "q", ["quiet", "exit-when-done"])
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err)
        return 1

    quiet = False
    exit_when_done = False
    for o, a in opts:
        if o in ('-q', '--quiet'):
            quiet = True
        elif o == '--exit-when-done':
            exit_when_done = True
        else:
            assert False, "unhandled option"

    if len(args) != 1:
        print >> sys.stderr, 'Exactly one configuration file must be provided.'
        return 1

    sinkmonitor = AdviseSinkMonitor()
    exitcontrol = ExitControl()
    tracker = ProcessTracker()

    try:
        cfg = config.Config(args[0])
    except config.ConfigException, ce:
        print >> sys.stderr, str(ce)
        return 1

    userlist = ({'user': x[0], 'server': x[1] % cfg.servers, 'sinkmonitor': sinkmonitor, 'exitcontrol': exitcontrol, 'connection': cfg.connection, 'bidirectional' : cfg.bidirectional} for x in cfg.mapping)
    requests = threadpool.makeRequests(tracker.start_migration, userlist, tracker.migration_completed, tracker. migration_exception)

    log_file = './zarafa-msr.log'
    if cfg.config.has_section('Logging'):
        sect_logging = dict(cfg.config.items('Logging'))
        check = config.checkSection(sect_logging.keys(), ['log_file'])
        if check:
            print >> sys.stderr, 'There are unknown options in the Logging section: %s' % str(check)
            return 1
        if 'log_file' in sect_logging:
            log_file = sect_logging['log_file']

    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)-8s %(name)s: %(message)s',
                        datefmt='%m-%d %H:%M',
                        filename=log_file,
                        filemode='a')

    if not quiet:
        console = logging.StreamHandler()
        console.setLevel(logging.INFO)

        formatter = logging.Formatter('%(levelname)-8s: %(message)s')
        console.setFormatter(formatter)

        logging.getLogger('').addHandler(console)

    logging.debug("Creating threadpool with %d workers", cfg.workers)
    pool = threadpool.ThreadPool(cfg.workers)

    logging.debug("Adding %u tasks to the threadpool" % len(requests))
    [pool.putRequest(req) for req in requests]
    pool.putRequest(threadpool.WorkRequest(sinkmonitor.Run, [pool], None))

    message_displayed = False
    while 1:
        try:
            time.sleep(0.5)
            pool.poll()
            if tracker.failed == len(requests):
                logging.critical("All migrations have failed, exiting!")
                sinkmonitor.Cancel()     # Cancel if all initial migrations failed
                break
            elif tracker.completed == len(requests):
                if message_displayed == False:
                    message_displayed = True
                    if exit_when_done:
                        logging.info("All migrations have completed successfully.")
                        exitcontrol.initiate_exit()
                        sinkmonitor.Cancel()
                        break
                    logging.info("All migrations have completed successfully, maintaining sync.")
            elif tracker.finished == len(requests):
                if message_displayed == False:
                    message_displayed = True
                    if tracker.failed > 0:
                        logging.warning("%u migrations have failed!" % tracker.failed)
                    if tracker.canceled > 0:
                        logging.warning("%u migrations were canceled!" % tracker.canceled)
                    if tracker.completed > 0:
                        logging.info("%u migrations have completed successfully, maintaining sync." % tracker.completed)

        except KeyboardInterrupt:
            print ""    # Get a newline after the ^C
            logging.info("User requested exit, please wait for shutdown...")
            exitcontrol.initiate_exit()
            sinkmonitor.Cancel()
        except threadpool.NoResultsPending:
            break

    logging.debug("Threadpool ready, waiting for workers to exit")
    pool.dismissWorkers(cfg.workers, True)
    logging.debug("Done")
    return 0
